/************************************************/
/*						*/
/*      Copyright:				*/
/*	 Jean-Marc Pigeon <jmp@safe.ca>	 2009	*/
/*						*/
/************************************************/
/* This program is free software; you can 	*/
/* redistribute it and/or modify it under the 	*/
/* terms of the GNU General Public License as	*/
/* published by the Free Software Foundation	*/
/* version 2 of the License			*/
/*						*/
/* This program is distributed in the hope that */
/* it will be useful, but WITHOUT ANY WARRANTY; */
/* without even the implied warranty of		*/
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR	*/
/* PURPOSE.  See the GNU General Public License	*/
/* for more details.				*/
/*						*/
/* You should have received a copy of the GNU	*/
/* General Public License along with this 	*/
/* program; if not, write to the Free Software	*/
/* Foundation, Inc., 51 Franklin Street,	*/
/* Fifth Floor, Boston, MA  02110-1301, USA.	*/
/************************************************/
/*						*/
/*	VZBOOT					*/
/*	Its purpose its too boot container 	*/
/*	process and wait for process completion	*/
/*	or a SIGTERM signal, to initiate the	*/
/*	container shutdown.			*/
/*						*/
/************************************************/
#include	<sys/wait.h>
#include	<sys/prctl.h>
#include	<sys/capability.h>
#include	<sys/utsname.h>
#include	<sys/personality.h>
#include	<errno.h>
#include	<sched.h>
#include	<stdio.h>
#include	<string.h>
#include	<syslog.h>
#include	<signal.h>
#include	<fcntl.h>
#include	"lowtyp.h"
#include	"dbgmem.h"
#include	"subprc.h"
#include	"subapl.h"
#include	"unilck.h"
#include	"unicnt.h"

#define	APPNAME	"vzgot"
#define	VZOPEN	"vzgot.open"	/*preparing VZ	*/
#define	VZCLOSE	"vzgot.close"	/*stopping VZ	*/
#define	VZSTART	"vzgot.start"	/*container init*/
#define	VZFBOOT	"vzgot.fboot"	/*container boot*/
#define	VZCREAT	"vzgot.create"	/*make container*/
#define	VZDEST	"vzgot.destroy"	/*undo container*/

#define	FBOOT	"firstboot"	/*firstboot rqst*/

/*argument structure				*/
typedef	struct	{
	pid_t mspid;	/*vzgot master PID	*/
	char *arch;	/*container architecture*/
	int argc;	/*number of argument	*/
	char **argv;	/*argument list		*/
	}ARGTYP;
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the program usage	*/
/*						*/
/************************************************/
static void usage()

{
(void) fprintf(stderr,"%s Version %s\n",appname,apl_getvers());
(void) fprintf(stderr,"usage:  %s "
		      "[-b base] "
		      "[-d debug] "
		      "[-f ] "
		      "[-h ] "
		      "[-v ] "
		      "action_word "
		      "container_name [starter]\n",appname);
(void) fprintf(stderr,"\t\t-b base      : application working directory\n");
(void) fprintf(stderr,"\t\t-d debug     : debug level\n");
(void) fprintf(stderr,"\t\t-f           : foreground\n");
(void) fprintf(stderr,"\t\t-h           : display usage\n");
(void) fprintf(stderr,"\t\t-v           : verbose debug\n");
(void) fprintf(stderr,"actionword       : boot|shutdown|firstboot|create|destroy,\n");
(void) fprintf(stderr,"container_name   : container name to do action to\n");
(void) fprintf(stderr,"[starter]        : program to start (default=\"/bin/init\")\n");
}
/*

*/
/************************************************/
/*						*/
/*	Procedure to display the firstboot usage*/
/*						*/
/************************************************/
static void fbusage()

{
(void) fprintf(stderr,"%s: %s Version %s\n",appname,FBOOT,apl_getvers());
(void) fprintf(stderr,"usage:  %s "
		      "[-D distname] "
		      "[-d domain] "
		      "[-i IPNUM,[IPALIAS,...]] "
		      "[-h ] "
		      "[-n nodename] "
		      "container_name [starter]\n",FBOOT);
(void) fprintf(stderr,"\t\t-l distname: container linux distribution\n");
(void) fprintf(stderr,"\t\t           : (default RedHat)\n");
(void) fprintf(stderr,"\t\t-d domain  : container domainmame\n");
(void) fprintf(stderr,"\t\t-h         : display firstboot usage\n");
(void) fprintf(stderr,"\t\t-b IPNUM   : container ipnum\n");
(void) fprintf(stderr,"\t\t-n nodename: container node name\n");
(void) fprintf(stderr,"container_name : container name to initialize\n");
(void) fprintf(stderr,"[starter]      : program to start\n");
(void) fprintf(stderr,"\t\t           : (default=\"/bin/init\")\n");
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to set the system architecture*/
/*						*/
/************************************************/
static int setarch(char *arch)

{
static	struct	{
		const char *arch;
		u_long persona;
		}pers[]={
			{"i386",PER_LINUX32},
			{"i686",PER_LINUX32},
			{"x86_64",PER_LINUX},
			{(const char *)0,PER_LINUX}
			};
int status;
int i;

status=-1;
errno=EINVAL;
for (i=0;pers[i].arch!=(const char *)0;i++) {
  if (strcmp(pers[i].arch,arch)==0) {
    status=personality(pers[i].persona);
    break;
    }
  }
return status;
}

/*
^L
*/
/************************************************/
/*						*/
/*	procedure to boot a container		*/
/*						*/
/************************************************/
static int doboot(void *args)

{
static int todrop[]={	/*capability to drop	*/
	CAP_SYS_BOOT,
	CAP_SYS_TIME
	};

#define	SBININIT	"/sbin/init"
ARGTYP *parms;
int status;
int foreground;
int phase;
int proceed;

/*SIGCONT received				*/
/*FIRST give time to master to send ITS SIGSTOP	*/
parms=(ARGTYP *)args;
status=0;
foreground=true;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*give time to master	*/
      (void) sleep(1);	/*to send SIGSTOP	*/
      break;		/*SIGCONT received 	*/
    case 1	:	/*lets protect host	*/
      if (todrop!=(int *)0) {	/*always	*/
	int i;

	for (i=0;i<sizeof(todrop)/sizeof(int);i++) {
          if (prctl(PR_CAPBSET_DROP,todrop[i],0,0,0)<0) {
	    (void) apl_alert(0,"%s, container %s, unable to remove "
			       "capability (todrop[%d]='%d', error=<%s>)",
			        appname,parms->argv[0],i,todrop[i],strerror(errno));
	    status=phase;
	    phase=999;	/*trouble trouble	*/
	    }
	  }
	}
      break;		/*SIGCONT received 	*/
    case 2	:	/*doing preliminary init*/
      if ((status=cnt_initscript(VZSTART,"%s %d %s",parms->argv[0],parms->mspid,base))!=0) {
	(void) apl_alert(0,"%s, container %s unable to init itself (return %d)",
			    appname,parms->argv[0],status);
	status=phase;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*doing root pivot	*/
      if (parms->argc==0)
         foreground=false;
      if (cnt_pivot(foreground,parms->argv[0])==false) {
	(void) apl_alert(0,"%s, container %s unable to pivot itself (aborting)",
			    appname,parms->argv[0]);
	status=phase;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 4	:	/*adjusting arch	*/
      if (setarch(parms->arch)<0) {
	(void) apl_alert(0,"%s, container %s unable to set architecture to <%s> (error=<%s>)",
			    appname,parms->argv[0],parms->arch,strerror(errno));
	status=phase;
	phase=999;	/*trouble trouble	*/
        }
      break;
    case 5	:	/*starting master jobs	*/
      parms->argc--;
      parms->argv++;
      if (parms->argc==0)	/*daemon start	*/
        (void) execl(SBININIT,SBININIT,(char *)0);
      else
        (void) execv(parms->argv[0],parms->argv);
      (void) apl_alert(0,"%s, container %s unable to start exec (errro=<%s>)",
			 appname,parms->argv[-1],strerror(errno));
      status=phase;	/*trouble with exec!	*/
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) sleep(1);	/*lets relax			*/
(void) exit(status);
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to stop a container		*/
/*						*/
/************************************************/
static int shutdown(int argc,char *argv[])

{
int status;
pid_t cntpid;
int phase;
int proceed;

status=1;
cntpid=(pid_t)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*getting container pid	*/
      if ((cntpid=cnt_getclonepid(argv[0]))==(pid_t)0) {
	(void) fprintf(stderr,"Shutdown is unable to find container <%s> "
			      "main pid\n",argv[0]);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*sending SIGINT	*/
      if (kill(cntpid,SIGINT)<0) {
	(void) fprintf(stderr,"Shutdown is send signal INT to container "
			      "<%s> main pid (error=<%s>)\n",
			      argv[0],strerror(errno));
        (void) cnt_rmclonepid(argv[0]);
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*waiting a little bit	*/
      (void) sleep(1);	/*container stop	*/
      break;
    case 3	:	/*waiting full stop	*/
      if (cntpid!=(pid_t)1) {	/*always	*/
	char *cmd;
	int i;

	(void) asprintf(&cmd," ps --ppid %d  | wc -l",cntpid);
	for (i=0;i<15;i++) {
	  FILE *fichier;
	  int remaining;

	  remaining=0;
	  if ((fichier=popen(cmd,"r"))==(FILE *)0) {
	    (void) apl_alert(0,"%s, shutdown pipe <%s> unable to proceed (error=<%s>)",
				appname,cmd,strerror(errno));
	    break;
	    }
	  if (fscanf(fichier,"%d",&remaining)!=1) 
	    (void) apl_alert(0,"%s, shutdown pipe <%s> not able to catch answer)",appname,cmd);
	  (void) fclose(fichier);
	  if (remaining<2) {
	    status=0;
	    break;	/*only remaining line "PID TTY TIME CMD"	*/
	    }
	  (void) sleep(1);
	  }
        (void) free(cmd);
	/*Terminating container main process	*/
  	(void) kill(cntpid,SIGKILL);
	}
      break;
    default	:	/*SAFE guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to create a container		*/
/*	structure.				*/
/*						*/
/************************************************/
static int create(int argc,char *argv[])

{
#define	NCREATE	4
#define	NPAR	"%s %s %s %s %s"

int status;
int phase;
int proceed;

status=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*check parameters num	*/
      if (argc<NCREATE) {
	(void) apl_alert(0,"Found %d parameter expecting %d",argc,NCREATE);
	status=1;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*calling initscript	*/
      if ((status=cnt_initscript(VZCREAT,NPAR,argv[0],argv[1],argv[2],argv[3],base))!=0) {
	(void) apl_alert(0,"Unable to create container <%s> (status=%d)",argv[0],status);
	status=2;
	phase=999;	/*trouble trouble	*/
        }
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to destroy a container	*/
/*	structure.				*/
/*						*/
/************************************************/
static int destroy(int argc,char *argv[])

{
#define	NDEST	1

int status;
int phase;
int proceed;

status=0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*check parameters num	*/
      if (argc<NDEST) {
	(void) apl_alert(0,"Found %d parameter expecting %d",argc,NDEST);
	status=1;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*calling initscript	*/
      if ((status=cnt_initscript(VZDEST,"%s %s",argv[0],base))!=0) {
	(void) apl_alert(0,"Unable to destroy container <%s> (status=%d)",argv[0],status);
	phase=999;	/*trouble trouble	*/
        }
      break;
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to start a container.		*/
/*						*/
/************************************************/
static int boot(int argc,char *argv[])

{
#define	STKSIZE	16384	/*clone own stack	*/

int status;
int cloneflg;
char *clonestk;
char *cntdist;
char *cntarch;
pid_t cpid;
int phase;
int proceed;

status=0;
cloneflg=SIGCHLD|__WCLONE;
cloneflg|=CLONE_NEWNET|CLONE_NEWIPC|CLONE_NEWNS|CLONE_NEWPID;
cloneflg|=CLONE_NEWUTS|CLONE_NEWUSER;
clonestk=(char *)malloc(STKSIZE);
cntdist=(char *)0;
cntarch=(char *)0;
cpid=(pid_t)0;
phase=0;
proceed=true;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*container name?	*/
      if ((argv[0]==(char *)0)||(strlen(argv[0])==0)) {
	(void) apl_alert(0,"No container name specified!");
	status=1;
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*checking container	*/
      if (cnt_iscontgood(argv[0])==false) {
	(void) apl_alert(0,"Aborting container start, check configuration!");
	status=phase;
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case 2	:	/*getting a lock	*/
      cntarch=cnt_getarch(argv[0]);
      cntdist=cnt_getdist(argv[0]);
      (void) prc_divedivedive(apl_foreground);
      (void) cnt_setstdio(apl_foreground,argv[0],"master.log");
      (void) prc_settitle("%s, starting %s container",appname,argv[0]);
      if (lck_locking(argv[0],LCK_LOCK,5)==false) {
	(void) apl_alert(0,"Unable to lock container <%s> access",argv[0]);
	status=phase;
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case 3	:	/*proceeding		*/
      if (argv[0]!=(char *)0) {	/*always	*/
	static ARGTYP args;
	
	args.mspid=getpid();
	args.arch=cntarch;
	args.argc=argc;
	args.argv=argv;
      	if ((cpid=clone(&doboot,(void *)(clonestk+STKSIZE),cloneflg,(void *)&args))<0) {
	  (void) apl_alert(0,"%s, Unable to clone %s container process (error=<%s>)",
			      appname,argv[0],strerror(errno));
	  status=phase;
	  phase=999;	/*trouble trouble	*/
	  }
	}
      break;
    case 4	:	/*call to init script	*/	
      /*stopping container first		*/
      (void) kill(cpid,SIGSTOP);
      if ((status=cnt_initscript(VZOPEN,"%s %d %d %s",argv[0],getpid(),cpid,base))!=0) {
	(void) apl_alert(0,"%s, Init script failure aborting %s container starting (return %d)",
			        appname,argv[0],status);
	(void) kill(cpid,SIGKILL);
	status=phase;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 5	:	/*setting clone PID	*/
      if (cnt_setclonepid(argv[0],cpid)==false) {
	(void) kill(cpid,SIGKILL);
	status=phase;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 6	:	/*waking up container 	*/
      if (kill(cpid,SIGCONT)<0) {
	(void) apl_alert(0,"%s, Unable to wakeup container %s process (error=<%s>)",
			    appname,argv[0],strerror(errno));
	(void) kill(cpid,SIGKILL);
	status=phase;
	phase=999;	/*trouble trouble	*/
	}
      break;
    case 7	:	/*waiting process	*/
      (void) prc_settitle("%s: container %s (%s/%s) up",appname,argv[0],cntdist,cntarch);
      (void) apl_settrap(true);
      (void) cnt_mstconsole(argv[0],cpid);
      (void) apl_settrap(false);
      cntdist=apl_freestr(cntdist);
      cntarch=apl_freestr(cntarch);
      break;
    default	:	/*SAFE guard		*/
      (void) prc_settitle("%s: container %s in shutdown process",appname,argv[0]);
      if ((status=cnt_initscript(VZCLOSE,"%s %d %s",argv[0],cpid,base))!=0) {
	(void) apl_alert(0,"%s, closing script failure for %s container stopping (return %d)",
			        appname,argv[0],status);
	}
      (void) cnt_rmclonepid(argv[0]);
      if (lck_locking(argv[0],LCK_UNLOCK,1)==false) {
	(void) apl_alert(0,"Unable to unlock container <%s> acces",argv[0]);
	status=10;
	}
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(clonestk);
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to start a container.		*/
/*						*/
/************************************************/
static int firstboot(int argc,char *argv[])

{
#define	DFLDIST	"RedHat"/*default distribution	*/

int status;
char *distrib;
char *domain;
char *ipnums;
char *nodename;
int phase;
int proceed;
char c;

status=0;
distrib=strdup(DFLDIST);
domain=(char *)0;
ipnums=(char *)0;
nodename=(char *)0;
phase=0;
proceed=true;
optind=1;
while ((c=getopt(argc,argv,"+D:d:hi:l:n:"))!=EOF) {
  switch(c) {
    case   'D'  :
      if (distrib!=(char *)0)
	(void) free(distrib);
      distrib=strdup(optarg);
      break;
    case   'd'  :
      if (domain!=(char *)0)
	(void) free(domain);
      domain=strdup(optarg);
      break;
    case   'i'  :
      if (ipnums!=(char *)0)
	(void) free(ipnums);
      ipnums=strdup(optarg);
      break;
    case   'n'  :
      if (nodename!=(char *)0)
	(void) free(nodename);
      nodename=strdup(optarg);
      break;
    case   'h'  :
    default	:
      (void) fbusage();
      break;
    }
  }
argc+=optind;
argv+=optind;
while (proceed==true) {
  switch (phase) {
    case 0	:	/*container name	*/
      if ((argv[0]==(char *)0)||(strlen(argv[0])==0)) {
	(void) apl_alert(0,"No container name specified!");
	status=1;
	proceed=false;	/*trouble trouble	*/
	}
      break;
    case 1	:	/*checking parameters	*/
      if (nodename==(char *)0)
	nodename=strdup(argv[0]);
      if (domain==(char *)0) {
	struct utsname un;
	char *ptr;

	(void) uname(&un);
	if ((ptr=strchr(un.nodename,'.'))!=(char *)0)
	  ptr++;
	else
	  ptr=un.nodename;
	domain=strdup(ptr);
	}
      if (ipnums==(char *)0)
	ipnums=strdup("");
      break;
    case 2	:	/*firstboot script	*/
      if ((status=cnt_initscript(VZFBOOT,"%s %s %s %s %s %s",
				 argv[0],distrib,domain,nodename,ipnums,base))!=0) {
	(void) apl_alert(0,"%s, container %s unable to proceed with %s (return %d)",
			    appname,argv[0],VZFBOOT,status);
	status=phase;
	phase=999;	/*trouble trouble	*/
	}
    default	:	/*SAFE Guard		*/
      proceed=false;
      break;
    }
  phase++;
  }
(void) free(nodename);
(void) free(ipnums);
(void) free(domain);
(void) free(distrib);
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	procedure to switch about proceeding	*/
/*						*/
/************************************************/
static int action(int argc,char *argv[])

{
int status;

status=0;
if (argc<=0) {
  (void) apl_alert(0,"%s action string requested");
  (void) usage();
  status=1;
  }
else {
  typedef enum	{
	a_firstboot,	/*container firstboot	*/
	a_boot,		/*booting container	*/
	a_shutdown,	/*container shutdown	*/
	a_create,	/*creating container	*/
	a_destroy,	/*removing container	*/
	a_unknown	/*unknown action	*/
	}ACTTYP;

  static VOCTYP vocaction[]={
	{a_firstboot,FBOOT,(const void *)0},
	{a_boot,"boot",(const void *)0},
	{a_shutdown,"shutdown",(const void *)0},
	{a_create,"create",(const void *)0},
	{a_destroy,"destroy",(const void *)0},
	{a_unknown,(char *)0,(const void *)0}
	};

  VOCTYP *voc;

  if ((voc=apl_getvoca(vocaction,argv[0]))!=(VOCTYP *)0) {
    switch(voc->code) {
      case a_firstboot	:	
	(void) strncpy(argv[0],voc->key,strlen(voc->key)+1);
        status=firstboot(argc,argv);
        break;
      case a_boot	:	
         status=boot(argc-1,argv+1);
        break;
      case a_shutdown	:	
        status=shutdown(argc-1,argv+1);
        break;
      case a_create	:	
        status=create(argc-1,argv+1);
        break;
      case a_destroy	:	
        status=destroy(argc-1,argv+1);
        break;
      case a_unknown	:
      default		:
        (void) apl_alert(0,"%s, <%s> unknown action string",appname,argv[0]);
        status=1;	
        break;
      }
    }
  }
return status;
}
/*
^L
*/
/************************************************/
/*						*/
/*	Program Entry.				*/
/*						*/
/************************************************/
int main(int argc,char *argv[])

{
int status;
char c;

appname=APPNAME;
status=0;
(void) apl_trapsegv(true);
(void) dbg_setcurpath(argv[0]);
memleak=DMTRACE;	/*setting memory leak detector ON	*/
(void) openlog(appname,LOG_PID,0);
base=strdup(base);
argv=prc_preptitle(argc,argv,environ);
while ((c=getopt(argc,argv,"+b:d:fh:v"))!=EOF) {
  switch(c) {
    case   'b'  :
      (void) free(base);
      base=apl_mkbase(optarg);
      break;
    case   'd'  :
      debug=atoi(optarg);
      break;
    case   'f'  :
      apl_foreground=true;
      break;
    case   'v'  :
      verbose=true;
      break;
    case   'h'  :
    default	:
      (void) usage();
      break;
    }
  }
(void) dbg_setcurbase(base);
if (argc>=optind) {
  status=action(argc-optind,argv+optind);
  }
else 
  (void) usage();
argv=prc_cleantitle();
(void) free(base);
(void) closelog();
(void) dbg_dumpmem(appname,"");
memleak=DMFALSE;	/*setting memory leak detector OFF	*/
(void) apl_trapsegv(false);
exit(status);
}
